// import 'dart:math';

import 'package:kenken/kenken.dart';
import 'package:test/test.dart';

void main() {
  group('Functions:', () {
    test('isUniq works right', () {
      expect(isUniq([1, 2, 3]), equals(true));
      expect(isUniq([1, 2, 1]), equals(false));
      expect(isUniq([1, 2, 2]), equals(false));
      expect(
          isUniq([
            1,
            2,
          ]),
          equals(true));
      expect(
          isUniq([
            1,
          ]),
          equals(true));
    });
  });

  group('Rule object', () {
    test('return cell to display rule in it', () {
      // Let create rule "5+" with such cells:
      // |-----------------------|
      // |      |        |       | # r - row
      // |------=================| # c - column
      // |      || r2,c2 | r2,c3 |
      // |---------------========|
      // |      || r3,c2 ||      |
      // |------==========-------|
      // In which cell to display rule "5+" ?
      var rule =
          Rule([RC(2, 2), RC(2, 3), RC(3, 2)], Operation(sum, label: '+'), 5);
      // Right answer is RC(2,2), because it is topmost and leftmost:
      var cellToDisplay = rule.coordinatesToDisplayTheRule;
      expect(cellToDisplay.row, equals(2));
      expect(cellToDisplay.col, equals(2));
    });
  });

  group('Kenken:', () {
    Kenken kenken;

    setUp(() {
      kenken = Kenken(3);

      kenken
        // RC 1,1 shoul be equal 1:
        ..addRule([RC(1, 1)], Operation(singleValue, label: '1'), 1)
        // Sum of cells 1,2 and 1,3 should be equal 5:
        ..addRule([RC(1, 2), RC(1, 3)], Operation(sum, label: '+'), 5)
        // Diff between 2,1 and 3,1 should be 1
        // Note: Using kenken.cellsOf() instead of RC()
        ..addRule(
            kenken.cellsOf([
              [2, 1],
              [3, 1]
            ]),
            Operation(diff, label: '-'),
            1)
        // sum of 2,2 2,3 should be 4
        ..addRule([RC(2, 2), RC(2, 3)], Operation(sum, label: '+'), 4);

      // Do not add rule below to kenken now.
      // Let kenken will be not fully covered by custom reles.
      // This rule we will add later in test.
      // sum of 3,2 and 3,3 should be 3:
      // ..addRule([RC(3, 2), RC(3, 3)], Operation(sum, label: '+'), 3);
    });

    test('`coordinatesOfRow` method returns right coordinates', () {
      var row2Coordinates = kenken.coordinatesOfRow(2);
      // first coordinate of row2Coordinates:
      expect(row2Coordinates[0].row, equals(2));
      expect(row2Coordinates[0].col, equals(1));
      // third coordinate of row2Coordinates:
      expect(row2Coordinates[2].row, equals(2));
      expect(row2Coordinates[2].col, equals(3));
    });

    test('rules not cover full kenken', () {
      expect(kenken.notCoveredByCustomRulesCells.length, equals(2));
    });

    test('rules cover full kenken', () {
      // Let add final rule.
      // sum of 3,2 and 3,3 should be 3:
      // Now the kenken is fully covered by custom rules.
      kenken.addRule([RC(3, 2), RC(3, 3)], Operation(sum, label: '+'), 3);
      expect(kenken.notCoveredByCustomRulesCells.length, equals(0));
    });

    test('Right solved kenken has not invalid rules', () {
      var values = [
        [1, 2, 3], // row = 1,
        [2, 3, 1], // row = 2,
        [3, 1, 2], // row = 3,
      ];

      // Fill kenken cells with values as it does user:
      for (var rowInd in List.generate(kenken.size, (i) => i)) {
        var row = rowInd + 1;
        for (var colInd in List.generate(kenken.size, (i) => i)) {
          var col = colInd + 1;
          kenken.setValue(row, col, values[rowInd][colInd]);
        }
      }
      // expect(awesome.isAwesome, isTrue);
      expect(kenken.brokenRules, equals([]));

      print('$kenken');
    });
  });
}
